/*

PS2MAdd Emulator for the SPEED ATA Controller patch loader

*/

#include <stdio.h>
#include <tamtypes.h>
#include <kernel.h>
#include <sifrpc.h>
#include <debug.h>
#include <loadfile.h>
#include <sio.h>

#include "p2mdp_common.h"

extern void	p2mesac_bin; //Generated by bin2s. 
extern u32	size_p2mesac_bin;

extern void atad_p2m_init_irx;
extern u32 size_atad_p2m_init_irx;

extern void ps2dev9_irx;
extern u32 size_ps2dev9_irx;

extern void poweroff_irx;
extern u32 size_poweroff_irx;



#define DPRINTF(args...) sio_printf(args)
#define DKPRINTF(...)  do { ee_kmode_exit(); EI(); DPRINTF(__VA_ARGS__); DI(); ee_kmode_enter(); } while(0)

//Reverse Word(32bit) Endianness
#define REWEN(x) ((((x) >> 24) & 0xFF) | (((x) & 0x00FF0000) >> 8) | (((x) & 0x0000FF00) << 8) | ((x) << 24))


//PPC instructions:

//		0	1	2	3	4	5		6	7	8	9	10		11	12	13	14	15		16	17	18	19	20	21	22	23	24	25	26	27	28	29	30	31

#define iBL(x)	(0x48000001 | ((x) & 0x03FFFFFC)) 
// 	0  1  0  0  1	0| 	LI																							AA	AK
//		18																														0	1

#define iNOP 	0x60000000 
// 	0  1  1  0  0	0| 	RS		 	|		RA			|				IM 									
//		24							0						0																	






void loadModules(void) {
	int ret;

	DPRINTF("Loading Modules\n");
	SifLoadFileInit();

	DPRINTF("Loading poweroff.irx\n");
	SifExecModuleBuffer(&poweroff_irx, size_poweroff_irx, 0, NULL, &ret);
	if (ret < 0) {
		DPRINTF("Failed loading poweroff.irx\n");
		return;
	}

	DPRINTF("Loading ps2dev9.irx\n");
	SifExecModuleBuffer(&ps2dev9_irx, size_ps2dev9_irx, 0, NULL, &ret);
	if (ret < 0) {
		DPRINTF("Failed loading ps2dev9.irx\n");
		return;
	}

	DPRINTF("Modules loaded.\n");
	return;
}

void loadAtadP2mInit(void) {
	int ret;
	DPRINTF("Loading atad_p2m_init.irx\n");
	SifExecModuleBuffer(&atad_p2m_init_irx, size_atad_p2m_init_irx, 0, NULL, &ret);
	if (ret < 0) {
		DPRINTF("Failed loading atad_p2m_init.irx\n");
		return;
	}
	return;
}


int main() {
	int i;

	SifInitRpc(0);

	DI();
	ee_kmode_enter();

	loadModules();

	DKPRINTF("Patching PPC-IOP RAM.\n"); 

	//One-time-initialization-code is run on the EE and then the configuration is communicated to the patch code. This is done through a shared structure.


	#define IOP_RAM2EE_ADDR(x) (((x) & ~UC_SEG_MASK) + IOP_RAM_BASE_ON_EE)
	#define IOP2DECK_LD_LUT_EE(x) (IOP_RAM2EE_ADDR(IOP2DECK_LD_LUT(x)))
	#define IOP2DECK_ST_LUT_EE(x) (IOP_RAM2EE_ADDR(IOP2DECK_ST_LUT(x)))


	//PPC_PATCH_ADDR is the address in PPC-IOP RAM where the patch code will be loaded. It comes from the main makefile. Edit the value in the makefile to change it.
	//PPC-IOP RAM executable segment is 0x00A00000 - 0x00AFFFFF.
	//It is accessible from EE side at 0xBCA00000 and at 0xBCE00000. 
	//IOP RAM (the entire 16MB of it) is accessible at 0xBC000000 on EE side. 

	#define IOP_RAM_BASE_ON_EE 0xBC000000
	#define BASEADDR (IOP_RAM_BASE_ON_EE + PPC_PATCH_ADDR)
	#define CODEADDR PPC_PATCH_ADDR

	//The addresses below would be used by the PPC-IOP ps2madd code.
	//The IOP hardware devices range on the PPC-IOP is available only at "segment" 0x1XXXXXXX.
	#define P2M_CMD_ADDR (0x1E000000)
	#define P2M_DAT_ADDR (0x1E000000)
	#define CDVD_STATUS_REG_ADDR (0x1F40200A)
	//Sets the ps2madd bus width P2M_BW8 / P2M_BW16.
	#define P2M_SET_BW P2M_BW16
	//Configure SSBUSC PIO channel DMA & Prg. I/O. //0x00003077; // 0x00003822; //0x00023822;
	#define P2M_SSBUS_CH_TM		0x00003022
	//0x00003010 is the fastest that works, but it is very unsafe.
	//#define P2M_SSBUS_CH_TM		0x00003010

	int spdStore = -1, spdLoad = -1, p2mdpInit = -1, dmac2Load = -1, dmac2Store = -1;
	u32 *DatPnt = &p2mesac_bin;
	struct sharedData *shDataP = (void*)-1;


	DKPRINTF(" sz= %08X \n", size_p2mesac_bin);
	DKPRINTF(" pnt= %08X \n", (u32)&p2mesac_bin);

	//Go through the patch file word-by-word, copying it to the PPC-IOP RAM.
	for (i=0; i<(size_p2mesac_bin / 4 + 1); i++) {
		*(vu32*)(BASEADDR + i*4) = DatPnt[i]; //Patch!
		if (shDataP == (void*)-1)
			if ((DatPnt[i] == REWEN(SHARED_MAGIC_0)) && (DatPnt[i+1] == REWEN(SHARED_MAGIC_1)))
				shDataP = (void*)(BASEADDR + i*4); //Map the shared structure.
	}

	if (shDataP == (void*)-1) {
		sio_putsn("ERR: Shared structure magic values were not found. Exiting.");
		//while(1);
		Exit(-1);
	}

	i = (u32)shDataP;
	DKPRINTF(" sharedDataPnt= %08X \n", i);
	i = REWEN((u32)shDataP->entry);
	DKPRINTF(" p2mdpInit addr= %08X \n", i);

	//Set configuration:
	shDataP->p2mDummyReg = REWEN(CDVD_STATUS_REG_ADDR);
	shDataP->p2mCmdRegAddr = REWEN(P2M_CMD_ADDR);
	shDataP->p2mDatRegAddr = REWEN(P2M_DAT_ADDR);
	shDataP->cfg0 = REWEN((P2M_CFG_BW & P2M_SET_BW)); //OR- other options inside.
	shDataP->p2mSsbusChTm = REWEN(P2M_SSBUS_CH_TM);

	//Get addresses of intercept functions:
	spdStore = REWEN((u32)shDataP->spdStore) - CODEADDR;
	spdLoad = REWEN((u32)shDataP->spdLoad) - CODEADDR;
	dmac2Load = REWEN((u32)shDataP->dmac2Load) - CODEADDR;
	dmac2Store = REWEN((u32)shDataP->dmac2Store) - CODEADDR;
	p2mdpInit = REWEN((u32)shDataP->entry) - CODEADDR;



	DKPRINTF(" Hooking Load Store handler functions.\n"); 
	//store:
	*(vu32*)(0xBCA0AA40) = REWEN(iBL((CODEADDR + spdStore) - 0xBCA0AA40));
	*(vu32*)(0xBCA0AA1C) = REWEN(iNOP); //A NOP which removes the branch in the SPEED store routine that makes ATA regs return only 0.
	//load:
	*(vu32*)(0xBCA0A978) =  REWEN(iBL((CODEADDR + spdLoad) - 0xBCA0A978));
	*(vu32*)(0xBCA0A94C) = REWEN(iNOP); //A NOP which removes the branch in the SPEED load routine that makes ATA regs return only 0.

	//DMA channel 8 - for Dev9 - SPEED - load and store handlers:
	*(vu32*)(0xBCA80144) = REWEN((CODEADDR + dmac2Load));
	*(vu32*)(0xBCAC0144) = REWEN((CODEADDR + dmac2Store));
	//For the DMAC 2 control regs (used for interrupts). May be better idea to use separate functions.
	*(vu32*)(0xBCA8015C) = REWEN((CODEADDR + dmac2Load));
	*(vu32*)(0xBCAC015C) = REWEN((CODEADDR + dmac2Store));


#if 0

	//Initialization function - called only once - it changes the address in the LUT, that is set to point to it here to point to a different function.
	//It is set here to intercept the IOP Interrupt controller range.
	//*(vu32*)(0xBCA8001C) = REWEN((CODEADDR + p2mdpInit));

	//The initialization function - for flushing the PPC cache (from PPC-side) and zeroing-out emulation registers is temporarely mapped to the address of 0xBF801070 - IREG (I_REG) - interrupt status reg. handler. Then the EE code (this) causes interrupt to IOP, but our handler function clears it before the MIPS-IOP driver handles it (it will see an empty IREG. This method doesn't work because of the PPC data cache, however, so triggering by loading from an unused address by IOP is used instead.

	u32 iopIntrMask = *(vu32*)(0xBF801074);
	*(vu32*)(0xBF801074) |= 0x02; //Enable the SBUS (PS1 GPU) interrupt.

	//cause the SBUS (PS1 GPU) intr.
	*(vu32*)(0xB000F240) &= ~0x00040000;
	int d = 0;
	while (d < 0x10) {d++;} //delay
	*(vu32*)(0xB000F240) |= 0x00040000; //interrupt to IOP.
	d = 0; //clear it back?
	while (d < 0x10) {d++;} //delay
	*(vu32*)(0xB000F240) &= ~0x00040000;

	//Now the init. func should be running.
	d=0;
	int e;
	DKPRINTF(" waiting for init func to start \n");

	while (*(vu32*)(0xBCA8001C) == REWEN((CODEADDR + p2mdpInit))) {
		d++;
		if (d>= 0xFFF) {
			DKPRINTF("\n ERR: init func doesn't seem to be running. \n"); 
			break;
		}
		e=0; while (e<0xFFFF) {e++;} //Give the func. some time - if we pool IOP memory too often, the PPC IOP won't have time to execute the function.
	}

	 *(vu32*)(0xBF801074) &= iopIntrMask | ~0x02; //Restore the setting of the PS1 GPU / intr by EE.

#endif



	//This makes loading from address INIT_LD_ADDR from IOP side cause the following address to be jumped to. On it, there is a routine which flushes the instruction cache. Witthout it, the new code won't get executed (correctly), because the changes in the SPEED load/store routines won't be seen by the CPU.
	//INIT_LD_ADDR (set in p2mdp_common.h) - The address which will be loaded from in the IOP module, so that the handler function mapped to it will be executed - p2mdpInit().

	(u32)(shDataP->initFuncLdLutOriginalValue) = *(vu32*)(IOP2DECK_LD_LUT_EE(INIT_LD_ADDR)); //save original value.
	 *(vu32*)(IOP2DECK_LD_LUT_EE(INIT_LD_ADDR)) = REWEN((CODEADDR + p2mdpInit));

	ee_kmode_exit();
	EI();

	//ps2atad has to be loaded separately, because if another driver accesses registers through the intercepting functions before the initialization function is called, the IOP will hang.
	loadAtadP2mInit();

	DI();
	ee_kmode_enter();
	int d = 0, e;

	while (*(vu32*)(IOP2DECK_LD_LUT_EE(INIT_LD_ADDR)) == REWEN((CODEADDR + p2mdpInit))) {
		d++;
		if (d >= 0xFFF) {
			DKPRINTF("\n ERR: Initialization function doesn't seem to have run. \n"); 
			break;
		}
		e=0; while (e < 0xFFFF) e++; //Give the function some time - if we pool IOP memory too often, the PPC IOP won't have time to execute the function.
	}
	ee_kmode_exit();
	EI();


	DPRINTF("All done!\n"); 
	Exit(-1);
	SleepThread();

}
