/*

P2MESAC: PS2MAdd Emulator for the SPEED ATA Controller
For the post-SCPH-75000 PS2 models - "DECKARD"-models.  

by Wisi 


*/



#include "types.h"
#include "p2mdp_common.h"



static int p2m_io_active = 0;
u32 dmaRunning;

u32 intrStat;
u32 intrMask;
u32 irqOverlay;


void updIntr(u32 intrMaskChange);

int p2m_io_start(void);
int p2m_io_finish(void);
void ataHddReset(int doRst);

u32 p2mdpInit(u32 addr, u32 type);
u32 emuSpdLoad(u32 offs, u32 type);
void emuSpdStore(u32 addr, u32 val, u32 type);
u32 dmac2LoadIntercept(u32 addr, u32 type);
void dmac2StoreIntercept(u32 addr, u32 val, u32 type);

void spdAtaCReset(void);

void dbgDelay(u32 per);
void dbg_putsS(char *str);
void prtHexS(u32 val);

//Comment-out the following line to disable the option.
//#define DEBUG_PRINT

#ifdef DEBUG_PRINT
	#define dbg_puts(x) dbg_putsS((x))
	#define prtHex(x) prtHexS((x))
#else
	#define dbg_puts(x) do{}while(0)
	#define prtHex(x) do{}while(0)
#endif




//Undefine this to make the bus-width settable through config from EE.
//#define FORCE_BUS_WIDTH P2M_BW16


static u16 P2M_FAST_RDH_8(const u8 reg);
static void P2M_FAST_WRH_8(const u8 reg, u16 var);
static u16 P2M_FAST_RDH_16(const u8 reg);
static void P2M_FAST_WRH_16(const u8 reg, u16 var);
//Pointers to one of the two pairs of functions above - according to bus width.
#ifndef FORCE_BUS_WIDTH 
	static u16 (*P2M_FAST_RDH)(const u8 reg);
	static void (*P2M_FAST_WRH)(const u8 reg, u16 var);
#else
	#if FORCE_BUS_WIDTH == P2M_BW16
		#define P2M_FAST_WRH(x) P2M_FAST_WRH_16((x))
		#define P2M_FAST_RDH(x) P2M_FAST_RDH_16((x))
	#else
		#define P2M_FAST_WRH(x) P2M_FAST_WRH_8((x))
		#define P2M_FAST_RDH(x) P2M_FAST_RDH_8((x))
	#endif
#endif



struct sharedData shData = {
	.magic[0] = SHARED_MAGIC_0,
	.magic[1] = SHARED_MAGIC_1,
	.entry = (void*)&p2mdpInit,
	.spdStore = (void*)&emuSpdStore,
	.spdLoad = (void*)&emuSpdLoad,
	.dmac2Load = (void*)&dmac2LoadIntercept,
	.dmac2Store = (void*)&dmac2StoreIntercept
};



//Hooks to the DECKARD code:

u32 (*comLoadHndlr)(u32 addr, u32 type) = (void*)0xA06BF4;
void (*comStoreHndlr)(u32 val, u32 addr, u32 type) = (void*)0xA06C34; //sub_A06C34

u32 (*intcLoad)(u32 addr, u32 type) = (void*)0x00A06D6C;
void (*intcStore)(u32 addr, u32 val, u32 type) = (void*)0x00A06E74;

//These functions handle loads & stores in the DMAC registers range 
u32 (*dmac1Load)(u32 addr, u32 type) = (void*)0x00A0789C;
void (*dmac1Store)(u32 addr, u32 val, u32 type) = (void*)0x00A07A18;
u32 (*dmac2Load)(u32 addr, u32 type) = (void*)0x00A07C7C;
void (*dmac2Store)(u32 addr, u32 val, u32 type) = (void*)0x00A07D44;




//Intercept functions:


//The intr. flag in the main intr. reg for DMAC 1 (bit3). 
#define IRQ_DMAC1 0x08
u32 irqDmac1Sv; //saved value of the above.

#define IRQ_VBLANK 0x01

u32 intrMaskSh;


u32 intcLoadIntercept(u32 addr, u32 type) {
	u32 val;
	val = intcLoad(addr, type);
	if (addr == 0x1F801070) { //I_STAT
		val |= irqOverlay;
	} else if (addr == 0x1F801074) { //I_MASK
		#if 1
		val &= ~IRQ_VBLANK & ~IRQ_DMAC1; //return only the mask the emulated IOP software set.
		val |= intrMaskSh;
		#else
		val = intrMaskSh;
		#endif
	}
	return val;
}

void intcStoreIntercept(u32 addr, u32 val, u32 type) {

	if (addr == 0x1F801074) {
		intrMaskSh = val; //Save the mask value the emulated IOP software sets.
		//val |= //whatever needs to be activated - done below.
	}

	if ((intrMask & 0x03) == 0) {
		if (addr == 0x1F801070) //I_STAT
			irqOverlay &= val; //clear non-set flags on write, but only if the driver ISR has cleared the mask.
	} else { //The driver is expecting an intr.
		if (addr == 0x1F801074) //an attempt to change/clear the mask
			val |= IRQ_VBLANK; //VBLANK is kept active, so that it would cause next intr. 
		//updIntr(0); //Check intr. and set status flags accordingly - this is called by the load handler for SPEED intrStat so it doesn't need to be called here. 
	}

	if (addr == 0x1F801074) { //an attempt to clear the mask
		//Save intr. flags that can get changed by this code.
		irqDmac1Sv = (irqDmac1Sv & ~IRQ_DMAC1) | (val & IRQ_DMAC1); 
		//val |= IRQ_DMAC1; //Keep the DMAC 1 intr. active. - shouldn't be necessary, unless emulation must be absolutely accurate ... is it used for the second DMAC intr. as well?
	}

	intcStore(addr, val, type); 
	return;
}



/*

From psx-spx:

Interrupt Acknowledge
To acknowledge an interrupt, write a "0" to the corresponding bit in I_STAT.
Most interrupts (except IRQ0,4,5,6) must be additionally acknowledged at the
I/O port that has caused them (eg. JOY_CTRL.bit4).
Observe that the I_STAT bits are edge-triggered (they get set only on
High-to-Low, or False-to-True edges). The correct acknowledge order is:
  First, acknowledge I_STAT                (eg. I_STAT.bit7=0)
  Then, acknowledge corresponding I/O port (eg. JOY_CTRL.bit4=1)
When doing it vice-versa, the hardware may miss further IRQs (eg. when first
setting JOY_CTRL.4=1, then a new IRQ may occur in JOY_STAT.4 within a single
clock cycle, thereafter, setting I_STAT.7=0 would successfully reset I_STAT.7,
but, since JOY_STAT.4 is already set, there'll be no further edge, so I_STAT.7
won't be ever set in future).


*/



//FIXME the stAddr is not really StartAddress!!! 
void flushDICaches(u32 stAddr, u32 endAddr) {
	__asm__ volatile ( \
		"dcbf\t%0,%1\n\t" \
		"sync\n\t" \
		"icbi\t%0,%1\n\t" \
		"isync\n\t" \
		: : "r" (stAddr), "r" (endAddr));
	return;
}

void invalDICaches(u32 stAddr, u32 endAddr) {
	__asm__ volatile ( \
		"dcbi\t%0,%1\n\t" \
		"sync\n\t" \
		"icbi\t%0,%1\n\t" \
		"isync\n\t" \
		: : "r" (stAddr), "r" (endAddr));
	return;
}


u32 msrSv;

inline void DIntr(void) {
	__asm__ volatile ( \
		"mfmsr\t%0\n\t" \
		"wrteei\t0\n\t" \
		: "=r" (msrSv));
	return;
}

inline void EIntr(void) {
	__asm__ volatile ( \
		"wrtee\t%0\n\t" \
		: : "r" (msrSv));
	return;
}


//Initialization function:
u32 p2mdpInit(u32 addr, u32 type) {

	//load: flush icache
	invalDICaches(0, 0x00A0A978);
	invalDICaches(0, 0x00A0A94C);

	//store: flush icache
	invalDICaches(0, 0x00A0AA40);
	invalDICaches(0, 0x00A0AA1C);

	//patches for LUT Load and Store DMA channel dev9 handlers flush icache
	invalDICaches(0, 0x00A80144);
	invalDICaches(0, 0x00AC0144);

	invalDICaches(0, 0x00A8015C);
	invalDICaches(0, 0x00AC015C);

	spdAtaCReset(); //initialize registers
	p2m_io_active = 0;
	//p2m_io_start();
	//ataHddReset(1); 

	intrMaskSh = intcLoad(0x1F801074, 4); //some initialization 

	//Configuration:
	#ifndef FORCE_BUS_WIDTH 
	if ((shData.cfg0 & P2M_CFG_BW) == P2M_BW8) { //8-bit bus
		P2M_FAST_RDH = P2M_FAST_RDH_8;
		P2M_FAST_WRH = P2M_FAST_WRH_8;
	} else { 
		P2M_FAST_RDH = P2M_FAST_RDH_16;
		P2M_FAST_WRH = P2M_FAST_WRH_16;
	}
	#endif

	*(vu32*)(IOP2DECK_LD_LUT(INIT_LD_ADDR)) = (vu32)(shData.initFuncLdLutOriginalValue);  //Restore to the original value.
	flushDICaches(0, IOP2DECK_LD_LUT(INIT_LD_ADDR));

	//Set the interceptors
	*(vu32*)(0x00A8001C) = (u32)&intcLoadIntercept;
	*(vu32*)(0x00AC001C) = (u32)&intcStoreIntercept;

	flushDICaches(0, 0x00A8001C);
	flushDICaches(0, 0x00AC001C);

	intcStore(0x1F801074, (intcLoad(0x1F801074, 4) & (~0x02)), 4); //mask (disable) 
	intcStore(0x1F801070, ~0x02, 4); //clear the interrupt flag (by wtiting 0 to the corresponding bit).


	dbg_puts("\nPS2MAdd patch for PPC-IOP initialized!\n");

	return 0; //Best don't access memory at all.
	//return shData.initFuncLdLutOriginalValue(addr, type);
	//return comLoadHndlr(addr, type);
	//return intcLoad(addr, type);
}




#define P2M_SSBUS_CH_TM (shData.p2mSsbusChTm)
// 0x00003022

//PS2MAdd-specific:

#define P2M_ctrl 0x40
#define P2M_stat P2M_ctrl    
//0x80 for read and 0 for write
#define P2M_RD_FLG	0x80
#define P2M_WR_FLG	0x00
//Send this as CMD, to switch the /CS for DATA to the second one (or to "anyOtherCs") (also used in emulated DMA).
#define P2M_DATA_CS	0x20
//When accessing the CPLD register(s), the HDD register select lines (addr. & /CS) can also be set to any desired value (which can be used to clear old states or for testing). 

//Bit switches in the CPLD register (P2MHW_ctrl - ps2madd HW Control reg.):
//The double clocking (on both DATA bytes) is not fixed yet (maybe it doesn't beeds to be) so write the same to both low and high byte. (high byte data is retained).
//When checked, after masking, checking must only be done by comparing with 0.
#define P2MC_ON         0x8080
#define P2MC_BUS_RST		0x4040
#define P2MC_END_COMM   0x2020
#define P2MC_TR_DIR     0x1010
#define P2MC_IRQ_STAT   0x0808
#define P2MC_IRQ_MASK   0x0404
#define P2MC_TR_MODE1   0x0202
#define P2MC_TR_MODE0   0x0101

// P2MC_IRQ_STAT is set when and while the HDD_IRQ line is act.low. It can be cleared (internal CPLD reg), by writing it back, but it won't be cleared if the HDD is keeping its HDD_IRQ line active low.  
//The satate-saving flip-flop was added, because with slower (delayed) access, the program sometimes misses the interupt.

#define ROM1_BASE_32 (*(vu32*)0x1FC00000)

#define P2M_EXTERN_DUMMY_REG (*(vu8*)(shData.p2mDummyReg))
#define TOGGLE_BUS_EXTEND (P2M_EXTERN_DUMMY_REG = 0)

#define P2M_CMD_BASE_ADDR (shData.p2mCmdRegAddr)
#define P2M_DAT_BASE_ADDR (shData.p2mDatRegAddr)

#define P2M_CMD_BASE 	(*(vu8*)P2M_CMD_BASE_ADDR)
#define P2M_DAT_BASE8 	(*(vu8*)P2M_DAT_BASE_ADDR)
#define P2M_DAT_BASE16 	(*(vu16*)P2M_DAT_BASE_ADDR)



static u16 P2M_FAST_RDH_8(const u8 reg) {
   u16 var;
DIntr();
	TOGGLE_BUS_EXTEND;  //added
	TOGGLE_BUS_EXTEND; //added 
   P2M_CMD_BASE = (reg | P2M_RD_FLG); 
	TOGGLE_BUS_EXTEND; 
	TOGGLE_BUS_EXTEND; 
   var = P2M_DAT_BASE8;
   TOGGLE_BUS_EXTEND; 
   var |= (P2M_DAT_BASE8 <<8) & 0xFF00;
   TOGGLE_BUS_EXTEND;//ADDED
EIntr();
   return var;
}

static void P2M_FAST_WRH_8(const u8 reg, u16 var) {
DIntr();
   TOGGLE_BUS_EXTEND;//ADDED
   P2M_CMD_BASE = (reg | P2M_WR_FLG); 
   TOGGLE_BUS_EXTEND;
   P2M_DAT_BASE8 = var;
   TOGGLE_BUS_EXTEND; 
	P2M_DAT_BASE8 = var >>8; 
	TOGGLE_BUS_EXTEND;//ADDED   
EIntr();
   return;
}


static u16 P2M_FAST_RDH_16(const u8 reg) {
   u16 var;
DIntr();
	//TOGGLE_BUS_EXTEND;
	TOGGLE_BUS_EXTEND;//here two were needed, but now it works with a single one as well.
   P2M_CMD_BASE = (reg | P2M_RD_FLG); 
	TOGGLE_BUS_EXTEND;
	TOGGLE_BUS_EXTEND; //here two are needed...
   var = P2M_DAT_BASE16;
   TOGGLE_BUS_EXTEND;
	//TOGGLE_BUS_EXTEND;
EIntr();
   return var;
}

static void P2M_FAST_WRH_16(const u8 reg, u16 var) {
DIntr();
	//TOGGLE_BUS_EXTEND;
   TOGGLE_BUS_EXTEND;
	P2M_CMD_BASE = (reg | P2M_WR_FLG); 
   TOGGLE_BUS_EXTEND;
   TOGGLE_BUS_EXTEND; //adding this one just in case
   P2M_DAT_BASE16 = var;
	TOGGLE_BUS_EXTEND;
	//-TOGGLE_BUS_EXTEND;
EIntr();
   return;
}

static inline u16 P2M_FAST_MOD(const u8 reg, u16 clrMask, u16 setMask) {
	//Reading back register value is no longer supported in ps2madd due to space constraints.
	//In case such is necessary, a shadow register should be used.
   u16 val = 0; //P2M_FAST_RDH(reg);
   P2M_FAST_WRH(reg, ((val & (~clrMask)) | setMask));
   return val;
}





//Debug printing:

void dbgDelay(u32 per) { //per in us
	int i;
	while (per > 0) {
		per--;
		i=0;
		while (i<217) i++; //(i<2176) i++;
	}
	return;
}


//For baud rate = 38400 bps:
#define DBG_BIT_PER 26

//FIXME: WARNING: Printing (this way) won't work while DMA transfer is active because of the way the PS2MAdd circuitry is made. 
void dbg_putc(char c) { //This does not contain p2m_io_start, so don't call it directly - use dbg_puts() instead.
	int i;
	DIntr();
	//For logic-level UART, the inactive (default state) of the output is high.
	//output is connected through an inverter.
	P2M_FAST_MOD(P2M_ctrl, 0, P2MC_TR_MODE0); //start bit = low on the line, which corresponds to high here.
	dbgDelay(DBG_BIT_PER);

	for (i=0; i<8; i++) { //send byte
		if ((c & 0x01) != 0)
			P2M_FAST_MOD(P2M_ctrl, P2MC_TR_MODE0, 0); //bit is = 1 => line goes high, this writes 0 to reg.
		else
			P2M_FAST_MOD(P2M_ctrl, 0, P2MC_TR_MODE0);
		dbgDelay(DBG_BIT_PER);
		c = c >> 1;
	}
	P2M_FAST_MOD(P2M_ctrl, P2MC_TR_MODE0, 0); //stop bit
	dbgDelay(DBG_BIT_PER*2);
	EIntr();
	return;
}

void dbg_putsS(char *str) {
	p2m_io_start();
	while (*str != 0)
		dbg_putc(*(str++));
	p2m_io_finish();
	return;
}

void sprthex(char *str, int inHex) {
	int i;
	i = 8;
	str[i] = 0; //Filling MSB to LSB
	i--;
	for (;;) {
		str[i] = inHex & 0xF;
		if (str[i] > 9) str[i]+=0x7;
		str[i]+=0x30;
		inHex = inHex >> 4; //4bits/nibble
		if (i == 0) break;
		i--;
	}
	return;
} 

void prtHexS(u32 val) {
	char str[10];
	sprthex(str, val); 
	dbg_puts(str);
	return;
}




void _entry(void) {
	return;
}


u32 xfrCtrl;
u32 dbufStat;
u32 ataDmaStat;
u32 ifCtrl;

u32 dmaCtrl;

u32 p2mDmaInterceptActive;

u32 dmac1PioOld;
u32 pioDmaChOld;
u32 dicr1Sv;



u32 RcommandCp;
u32 RfeatureCp;
u32 RnsectorCp;
u32 RsectorCp;
u32 RlcylCp;
u32 RhcylCp;
u32 RselectCp;


#if 1
u32 setTransferModeIntercept(u32 adr, u16 val) {

	#define ATA_C_SET_FEATURES 0xEF

	if (adr == 0x0E) { //r_command register
		RcommandCp = val & 0xFF;
		if (RcommandCp == ATA_C_SET_FEATURES) {
			if ((RfeatureCp == 0x03) && (RsectorCp == 0x0) && (RlcylCp == 0x0) && (RhcylCp == 0x0)) { //it is a transfer-mode setting command
			#if 1
				if ((RnsectorCp & 0x68) == 0x08) { //mode is PIO
					P2M_FAST_WRH(0x0A, (0x08 | 0x04)); //write new data to reg. 0x04 for mode = PIO(0x8),4(0x4)
					P2M_FAST_WRH(0x0A, (0x08 | 0x04)); //write twice in case of 48-bit LBA
				} else {
					P2M_FAST_WRH(0x0A, (0x20 | 0x02)); //write new data to reg. 0x04 for mode = MDMA(0x20),2(0x2)
					P2M_FAST_WRH(0x0A, (0x20 | 0x02));
				}
			#else
				P2M_FAST_WRH(0x0A, RnsectorCp); //write twice in case of 48-bit LBA
				P2M_FAST_WRH(0x0A, RnsectorCp);
				P2M_FAST_WRH(0x0E, RselectCp);
			#endif
			}
		}
	} else if (adr == 0x02) { //r_feature
		RfeatureCp = val;
	} else if (adr == 0x04) { //r_nsector
		RnsectorCp = val;
	} else if (adr == 0x06) { //r_sector //basically these below can be removed
		RsectorCp = val;
	} else if (adr == 0x08) { //r_lcyl
		RlcylCp = val;
	} else if (adr == 0x0A) { //r_hcyl
		RhcylCp = val;
	} else if (adr == 0x0C) { //r_select
		RselectCp = val;
	}
	return val;
}
#endif 


u16 ataRegLoad(u32 offs) {
p2m_io_start();
	offs = ((offs - 0x40) >>1) ^ 0x08;
	if ((offs & 0x08) == 0) offs |= 0x10; //my mapping is weird...
	//probably it would be a good idea to re-wire PS2MAdd to use closer mapping (shift all bits one left).
	//	reg		0	2	4	6	8	A	C	E
	//	cs0		08	09	0A	0B	0C	0D	0E	0F
	// cs1		10	11	12	13	14	15	16	17
	u16 val = P2M_FAST_RDH(offs);
p2m_io_finish();
	return val;
}


void ataRegStore(u16 val, u32 adr) {
p2m_io_start();
	u32 offs = ((adr - 0x40) >>1) ^ 0x08;
	if ((offs & 0x08) == 0) offs |= 0x10; 
	setTransferModeIntercept((adr-0x40), val);
	P2M_FAST_WRH(offs, val);
p2m_io_finish();
	return;
}



int p2m_io_start(void) {
	if (p2m_io_active)
		return p2m_io_active; 
//	if (ROM1_BASE_32 == 0x401a7800) // check the 1st in ROM0 seg get
	p2m_io_active = 0;
	// PS2MAdd start commands sequence 
DIntr();
	P2M_CMD_BASE = 0x72;
	TOGGLE_BUS_EXTEND;
	P2M_CMD_BASE = 0x34;
	TOGGLE_BUS_EXTEND;
	P2M_CMD_BASE = 0x61;
	TOGGLE_BUS_EXTEND;
	//#if HW_TYPE == PS2MADD_CLONE
		TOGGLE_BUS_EXTEND;	//HW needs more time for reason unknown.
	//#endif
	unsigned int res = P2M_CMD_BASE;
	TOGGLE_BUS_EXTEND;
EIntr();
	if ((res & 0xff) == 0xe7) 	// check result
		p2m_io_active = 1;
	return res;
}


int p2m_io_finish(void) {
	if (p2m_io_active == 0) return 0;
	if (dmaRunning == 1) return 0; //don't end connection over dma
	//if (p2m_io_active == 1)
		P2M_FAST_MOD(P2M_ctrl, 0, P2MC_END_COMM); //set end flag (it is auto-cleared). 
	//for (i=0; i<0xFFFFF; i++); //there shouldn't be a need for delay here.
//	if (ROM1_BASE_32 == 0x401a7800) // check the 1st in ROM0 seg get
		p2m_io_active = 0;	 // back to it's original state
	return p2m_io_active ^ 1;
}


/*
revision:

REV_3 is kept at 0xBE0798 in RAM. REV_8 is hard-coded in the load routine to 0x2.
Basically no change to the rev. regs should be necessary.

*/



void spdAtaCReset(void) {
	//This function initializes variables used for emulation.
	//It should instead be a reset function for the virtual SPEED ATA Controller (as the name suggests)... 

	xfrCtrl = 0;
	dbufStat = 0;
	ataDmaStat = 0;
	ifCtrl = 0;

	intrStat = 0;
	intrMask = 0;
	dmaCtrl = 0;

	p2mDmaInterceptActive = 0;
	dmaRunning = 0;

	dmac1PioOld = 0;
	pioDmaChOld = 0;

	dicr1Sv = 0;

	irqOverlay = 0;

	RcommandCp = 0;
	RfeatureCp = 0;
	RnsectorCp = 0;
	RsectorCp = 0;
	RlcylCp = 0;
	RhcylCp = 0;
	RselectCp = 0;

	return;
}



/*
Because the (ps2atad and ps2dev9) drivers running on the emulated IOP rely on SPEED and DMA interrupts in some cases, this program needs a mechanism to cause an interrupt.

For the SPEED ATA0 interrupt, another interrupt cause (VBLANK) is activated and used to trigger the main interrupt dispatcher. From there, this code emulates the main interrupt status register and the SPEED interrupt registers, effectively faking that a SPEED ATA interrupt has occured. On loading from the SPEED interrupt status register, updIntr() is called, which checks the IRQ line from the HDD, and sets the corresponding flag in the SPEED interrupt status register if interruot has occured. Basically the ATA interrupt is pooled on each VBLANK. This makes transfers slower, when the HDD needs to seek more.

For the SPEED ATA1 interrupt - activated when data is available from HDD in the SPEED buffer or when the SPEED buffer can accept data from IOP, the VBLANK source is simply used to cause it as soon as possible. This is done because PS2MAdd has no internal buffer, so as soon as the HDD is ready to send or receove data, PS2MAdd is ready as well. For this reason the SPEED interrupt status register falg for this interrupt is kept always set.

The Dev9 DMA interrupt (used by some versions of ps2dev9) is faked by the PIO DMA channel interrupt (the PIO channel of the SSBUSC is used for DMA transfers with PS2MAdd). Because the PIO and the Dev9 DMA channels are part of separate DMACs, emulating the registers' behaviour is too complex, so only the most necessary parts are emulated.

Only the some (a few) of the SPEED's ATA controller registers are emulated, because the rest are rarely used. The homebrew ps2atad drivers work correctly with this emulator, however the drivers in PS2Linux for example most likely won't work. Also, because PS2Linux accesses the SPEED hardware from EE side, the accesss won't go through this program at all.  


Waiting for interrupt: The dev9 intr. is kept active (at the dev9 itself as well). Some intr. is used to cause an intr. on some (fixed) intervals (VBLANK). Once the main intr. to the IOP is triggered, the main ISR will go over all sources and jump to the dev9, which is emulated by this code as active. Then the dev9 ISR (in the dev9 driver) will go through the SPEED intr. STAT and MASK regs and look for matching bits. This of itself is a read of the STAT reg, which checks the HDD interrupt line.  

To create a (very inefficient) "pooling" effect, once the SPEED intrMask is activated (intrEnable), the intr. handler here causes the global intr, (VBLANK is used), but does not set any of the SPEED intrStat bit-fields, (unless the intr condition did occur) so that the SPEED ISR will exit, keeping the intr. active (an isr won't be called). On this condition, the main intr. causing source is kept active, so that the intr will continue to trigger, untill thie condition is met.

Here is what should happen:
1. driver enables intr. through its mask reg.  (the event flag was cleared prior to that), the dev9 intr is enabled by the dev9 driver by default (always). 
2. This emulator captures the writing to the speed intrMask reg and when it detects that a bit has been set (bits 0 and/or 1) it enables the VBLANK intr. and sets the shadow reg (intrOverlay) for the dev9 intr., so that it would appear as active to the main ISR. It also sets bits in the intrStat shadow reg in case the intr cause is active.
3. Now that (while) intrMask (shadow) has bits (0,/1) set, the main intr regs store interceptor will keep intrOverlay set and will keep the mask bit for VBLANK active, so that on the next VBLANK the intr will occur again.
4. When the main ISR (triggered by the VBLANK intr cause) detects the Dev9 bit set, it will call its handler as well. The Dev9 isr will read the speed intrStat and intrMask regs. The intrStat load interceptor will check the intr. condition(s) and set the intrStat shadow accordingly. It will also re-trigger the VBLANK intr if the intrMask shadow has bits set.
5. If the interrupt condition has occured, the instrStat would be changed to reflect that and the driver ISR will see that and clear the intrMask, completing this. If the condition is still not met, the dev9 isr won't call the corresponding handler, the intrMask shadow would remain set and any attempts to cler the VBLANK intr. mask or the intrOverlay will not succeed. This will loop back.

What could cause problems, is that the driver may set-up the DMA channel before setting-up the (virtual) SPEED ATA Controller. This would cause the DMA transfer to be run to the SMAP instead and more problems to happen (this code would be left waiting for a transfer).
Furthermore we have no way of knowing (in advance) whether a SMAP or ATA DMA transfer is starting, by just the SPEED regs. It seems best to constantly monitor both the DMAC and the SPEED regs., and once the conditions for an ATA transfer are met, to configure the DMAC accordingly and start it.  


*/
 



void causeIntr(u32 iMask) {
	if ((iMask & 0x03) != 0) { //
		irqOverlay = 0x00002000; //set the intr. flag., which is checked by the main ISR (intrman). It isn't used by the dev9 driver.
		//trigger intr. to emulated IOP: 
		//trigger by causing some other intr.
		intcStore(0x1F801074, (intcLoad(0x1F801074, 4) | 0x01), 4); //misuse the vblank interrupt, which hapens commonly anyway
	} else {
		irqOverlay = 0;
		//Deactivate the mask only if the emulated IOP code doesn't use it.
		if ((intrMaskSh & IRQ_VBLANK) == 0)
			intcStore(0x1F801074, (intcLoad(0x1F801074, 4) & ~0x01), 4); 
	}
	return;
}


void updIntr(u32 intrMaskChange) {
	//intrMask here will contain the value from before the write to it (if intrMaskChange != 0)
	if (intrMaskChange & 0x02)
		if (((intrMaskChange ^ intrMask) & 0x02) == 0) { //the dma fifo intr is inact
			intrStat &= ~2;
		}
p2m_io_start();
	//Clear the ps2madd internal intr. latch each time the mask is changed (or only when cleared).
	if (intrMaskChange & 0x01) { //ATA0 intr - for ATA - the INTRQ line
	//	if ((intrMask & 0x01) == 0)
			P2M_FAST_MOD(P2M_ctrl, P2MC_IRQ_MASK, P2MC_IRQ_STAT); //clear interrupt (by wting 1 to IRQ_STAT). Also disables (only) the hardware IRQ line by claearing IRQ_MASK.
		intrStat &= ~1; //Clear the intrStat shadow here as well, to prevent missing an interrupt if cleared at the wrong time - the intr. reg. of ps2madd could get cleared too early, (the hdd would report so) and on the second read of intrStat (or more) it could be left clear. 	
	}
	//read intr. from HDD:
	if (P2M_FAST_RDH(P2M_ctrl) & P2MC_IRQ_STAT)
		intrStat |= 1;

p2m_io_finish();

	causeIntr(intrMaskChange ^ intrMask);
	return;
}







void ataHddReset(int doRst) {
p2m_io_start();

	if (doRst == 0) { //0 => do reset
	   P2M_FAST_MOD(P2M_ctrl, 0, P2MC_BUS_RST | P2MC_END_COMM); //Setting the bit, resets the bus (brings HDD_RST low). 
	} else {
		P2M_FAST_MOD(P2M_ctrl, P2MC_BUS_RST, P2MC_END_COMM); //get line inact.high
	}
	p2m_io_active = 0;
//p2m_io_finish(); // Placed above, directly for obvious reasons - ending ps2madd io will turn hdd on (exit reset), if the reset causing bit is not set in the same register access.
	return;
}


//Normally the driver will only start the DMA transfer (DMAC) when it has received an interrupt.
//This function is not really used.
void p2mRunDma(void) {
	if ((xfrCtrl & 0x80) == 0) return;
//	if ((xfrCtrl & 0x80) == 0) return;
	//if ((ifCtrl & 0xC0) != 0) return; //reset switches - not really necessary 
	//if HDD is ready ...  / wait until it becomes ready
	if ((ifCtrl & 0x04) == 0) return; //auto-cleared at end of transfer
	if ((dmaCtrl & 0x01) == 1) return; // target is SMAP => exit
/*
	if (dmacReady) {
	} */
	ifCtrl &= ~0x04;
	return;
} 



#define DICR2_DEV9_EN	0x00020000
#define DICR2_DEV9_FLG	0x02000000

#define DICR_PIO_EN	0x00200000
#define DICR_PIO_FLG	0x20000000



inline void enableP2mDmaChCfg(void) { //Configures a DMA channel, chosen for PS2MAdd.
	dmac1PioOld = dmac1Load(0x1F8010F0, 4);
	pioDmaChOld = *(u32*)0x1F80101C;

	// *(u32*)0x1F8010F0 |= 0x00800000;
	dmac1Store(0x1F8010F0, ((dmac1Load(0x1F8010F0, 4) & 0xFF0FFFFF) | 0x00800000) , 4);
	*(u32*)0x1F80101C = ((*(u32*)0x1F80101C) & 0x001F0000 ) | P2M_SSBUS_CH_TM; // 0x00003077; // 0x00003822; //configure SSBUSC PIO channel DMA & Prg. I/O. This can (should) be done from EE, so it is fine here.
	return;
}


inline void restoreP2mDmaChCfg(void) { //Restores the original configuration of the DMA channel, chosen for PS2MAdd.
	dmac1Store(0x1F8010F0, ((dmac1Load(0x1F8010F0, 4) & 0xFF0FFFFF) | (dmac1PioOld & 0x00F00000)) , 4);
	*(u32*)0x1F80101C = pioDmaChOld;
	return;
}


inline void enableP2mDmaIntrCfg(void) {
	dicr1Sv = dmac1Load(0x1F8010F4, 4);
	//intrMaskSv = *(u32*)0x1F801074;
	dmac1Store(0x1F8010F4, ((dmac1Load(0x1F8010F4, 4) & 0x00FFFFFF) | DICR_PIO_EN | 0x00800000), 4);	 //and enabl master flag
	*(u32*)0x1F801074 |= 0x8; //Enable the DMAC1 interrupt source in the global interrupt register.
	return;
}

#if 0
//This is not used, because without intercepting the DMAC1 loads and stores, it can't be guaranteed that the emulated IOP won't change its configuration and this function won't write some old value afterwards.
inline void restoreP2mDmaIntrCfg(void) {
	dmac1Store(0x1F8010F4, dicr1Sv, 4);
	//*(u32*)0x1F801074 intrMaskSv;
	return;
}
#endif


//For now assuming that the DMA transfer is set-up and started after configuring the SPEED for DMA.

u32 dmac2LoadIntercept(u32 addr, u32 type) {
	u32 val = 0;
	u32 dmy;

	if (addr == 0x1F801574) {
		val = dmac2Load(addr, type);
		if ((dmac1Load(0x1F8010F4, 4) & DICR_PIO_FLG) != 0) { //The PIO DMA channel interrupt has triggered => fake that the Dev9 DMA channel interrupt has occured.
			val |= DICR2_DEV9_FLG;
		/*	dbg_puts("\r\n DMAC2 intr Load = "); //Printing this way over DMA doesn't work because the logic circuit can't receive such commands.
			prtHex(val);
			dbg_puts(" ");  
		*/
			if ((dmaRunning == 1) && ((dmac1Load(0x1F8010D8, 4) & 0x01000000) == 0)) {
DIntr();
					dmaRunning = 0;
					//DMA was running until now and just ended => end ps2madd DMA mode
					if ((xfrCtrl & 1) == 0)
						dmy = P2M_CMD_BASE; //Advance the state machine to end the transfer F1=1;F2=0 => F1=0;F2=0
					else
						P2M_CMD_BASE = 0;
					TOGGLE_BUS_EXTEND;    
					/*dmy = P2M_CMD_BASE; //Another (this) one will be necessary to for 8-bit mode. 
					TOGGLE_BUS_EXTEND;     */
					P2M_FAST_MOD(P2M_ctrl, (P2MC_TR_MODE1 | P2MC_ON), 0); //end
p2m_io_finish();
EIntr();
					dbg_puts("\r\n DMA End3 ");
			}
			return val;
		}

/*
	if ((dmaRunning == 1) && ((dmac1Load(0x1F8010D8, 4) & 0x01000000) == 0)) {

			val |= DICR2_DEV9_FLG;
	dbg_puts("\r\n DMAC2 intr Load = ");
	prtHex(val);
	dbg_puts(" "); 
dbgDelay(1000000);
goto finishDma;

		}
*/
	//it is possible that by the time the DMAC2 intr reg is read, the PIO dma intr in DICR1 would be cleared already.
	}

	if (p2mDmaInterceptActive) {
		if ((addr & 0xFFFFFFF0) == 0x1F801510) { //Dev9 - SPEED
			val = dmac1Load((0x1F8010D0 | (addr & 0x0F)), type);
			if ((addr & 0x0F) == 0x08) { //CHCR
				if ((dmaRunning == 1) && ((val & 0x01000000) == 0)) {
//finishDma: 
					dbg_puts("\r\n DMA End ");
DIntr();
					dmaRunning = 0;
					//DMA was running until now and just ended => end ps2madd DMA mode
					//if ((val & 1) == 0) //read
					if ((xfrCtrl & 1) == 0) //read //Taking the direction from xfrCtrl instead (might be unsafe).
						dmy = P2M_CMD_BASE; //Advance the state machine to end the transfer F1=1;F2=0 => F1=0;F2=0
					else
						P2M_CMD_BASE = 0;
					TOGGLE_BUS_EXTEND;    
					/*dmy = P2M_CMD_BASE; //Another (this) one will be necessary to for 8-bit mode. 
					TOGGLE_BUS_EXTEND;     */
					P2M_FAST_MOD(P2M_ctrl, (P2MC_TR_MODE1 | P2MC_ON), 0); //end
p2m_io_finish();
EIntr();
					restoreP2mDmaChCfg();
					dbg_puts("\r\n DMA End2 ");
				}
			}
			/*
			//This can't be done, because there is nothing to enable intercepting the next transfer.
			//Best bet is to end the transfer either on ATA DMA regs in SPEED disable or (and) DMA target change (to SMAP).
			if ((val & 0x01000000) == 0) { //transfer was completed
				p2mDmaInterceptActive = 0;
			}
			*/
			return val;
		}
	}
	return dmac2Load(addr, type);
}


void dmac2StoreIntercept(u32 addr, u32 val, u32 type) {
	
	//For now only keep the PIO intr always enabled and only take care of clearing it.
	if (addr == 0x1F801574) {
		if (val & DICR2_DEV9_FLG) { //The driver is trying to clear the intr. flag => clear the PIO intr flag as well 
			dmac1Store(0x1F8010F4, ((dmac1Load(0x1F8010F4, 4) & 0x00FFFFFF) | DICR_PIO_FLG), 4);	
		}
		//The following is done below instead, because it won't get executed here.
		/*if (val & DICR2_DEV9_EN) { //the driver has enabled the dev9 DMA intr => enable the PIO DMA intr.
			enableP2mDmaIntrCfg();
		} */
	}

	if (p2mDmaInterceptActive) {
		if ((addr & 0xFFFFFFF0) == 0x1F801510) { //Dev9 - SPEED
			if ((addr & 0x0F) == 0x08) { //CHCR
				val = (val & 0x01000201) | 0x00000200; //
				//Here - in ps2madd "real" DMA mode, the channel is NOT force-started, so the configuration the driver uses should work fine.	
				if ((val & 0x01000000) && (dmaRunning == 0)) { //DMA start
					dbg_puts("\r\n DMA Start ");
					//Enable the PIO (in DMAC1) intr.
					//This is not done above, because by the time this patch is loaded, the Dev9 driver might already have been loaded, and only on initialization does it enable the Dev9 DMA channel interrupt. Also because the original state is restored after DMA, it has to be set each time DMA starts as well (here).
					enableP2mDmaIntrCfg();
					//if (dmaRunning == 0) {
					enableP2mDmaChCfg();
					//}

p2m_io_start();
DIntr();
					dmaRunning = 1;
					P2M_FAST_MOD(P2M_ctrl, 0, (P2MC_TR_MODE1 | P2MC_ON)); //DMA set-up & start
				if ((val & 0x01) == 0) //read
					P2M_CMD_BASE = (P2M_RD_FLG | P2M_DATA_CS); //Start (enable) transfer (read). 
				else
					P2M_CMD_BASE = (P2M_WR_FLG | P2M_DATA_CS);
				TOGGLE_BUS_EXTEND;     //This is not entirely necessary.
				//... and now hope that nothing would try to access ps2madd or the used address range or the PIO DMA channel at this point.
EIntr();
//ps2mad I/O is disabled only once the DMA transfer ends, in dmac2StoreIntercept().
				}
			}
			dmac1Store((0x1F8010D0 | (addr & 0x0F)), val, type);	//write to the PIO DMA channel regs.
			return;
		}
	}

	dmac2Store(addr, val, type);
	return;
}



u32 emuSpdLoad(u32 addr, u32 type) {
	u32 offs;
	u32 val = 0;
	offs = addr & 0xFFFF; //only SPEED range. 
	//Addresses are sent to this routine with base addr. = 0x13000000. This is where the SPEED really is in the PPC IOP mapping.


if ((offs >= 0x20) && (offs <= 0x7F)) {
	dbg_puts("\r\n SpdLd ");
	prtHex(offs);
	dbg_puts(" = ");
}

	if ((offs >= 0x40) && (offs <= 0x5E)) { //ATA 
		val = ataRegLoad(offs);
	}


	switch (offs) {
		case 0x24: //	SPD_R_DMA_CTRL	
			val = comLoadHndlr(addr, type); //might need to add the ATA switch flag.
			break;
		case 0x28: //	SPD_R_INTR_STAT
			updIntr(0);
			val = ((comLoadHndlr(addr, type) & 0xFFFC) | intrStat);
			break;
		case 0x2A: //	SPD_R_INTR_MASK
			val = ((comLoadHndlr(addr, type) & 0xFFFC) | intrMask);
			break; 
		case 0x32: //	SPD_R_XFR_CTRL
			val = xfrCtrl;
			break;
	/*	case 0x34: //	Total number of bytes, read from SPEED
//For now don't emulate these, even though PS2Linux uses them ...
			break;
		case 0x36: //	Total number of bytes, writen to SPEED

			break; */
		case 0x38: //	SPD_R_DBUF_STAT
			val = 0x3F; //0x21; //dbufStat;
			//Using large values (0x1F) might break stuff if atad really trusts this.
			break;
		//skipping the 0x60,... - weird registers.

		case 0x62: //	State of (Ultra ?) DMA transfer and lines.
			val = ataDmaStat;
			break;
		case 0x64: //	SPD_R_IF_CTRL
			val = ifCtrl;
			break;
		/*case 0x66: // r/w? it is writtable up to 0xFF and usually contains 0x1
			return spdReg0x66;
			break; */

//Might as well not emulate these:
/*		case 0x70: //	SPD_R_PIO_MODE
			return pioTiming;
			break;
		case 0x72: //	SPD_R_MWDMA_MODE
			return mwdmaTiming;
			break;
		case 0x74: //	SPD_R_UDMA_MODE
			return udmaTiming;
			break;
		case 0x76: //	CRC Initial value 
			return crcIn;
			break;
		case 0x78: //	CRC result 
			return crcOut;
			break; */

		default: 
			if ((offs <= 0x30) || (offs >= 0x80)) //This is done by the original code, that calls this function.
				val = comLoadHndlr(addr, type); //common handler
	}


	if ((offs >= 0x20) && (offs <= 0x7F)) {
		prtHex(val);
		dbg_puts("  ");
	}

	return val;
}




void emuSpdStore(u32 addr, u32 val, u32 type) {
	u32 offs = addr & 0xFFFF;
	u32 tmp;
	if ((offs >= 0x20) && (offs <= 0x7F)) {
		dbg_puts("\r\n SpdSt ");
		prtHex(offs);
		dbg_puts(" = "); 
		prtHex(val);
		dbg_puts(" "); 
	}

	if ((offs >= 0x40) && (offs <= 0x5E)) { //ATA
		ataRegStore(val, offs);
		//return;
	} else {

	switch (offs) {
		case 0x24: //	SPD_R_DMA_CTRL	
			dmaCtrl = val;
			p2mDmaInterceptActive = (~val) & 0x01; //activate DMA-channel set-up when the target is switched.
			//if ((val & 0x01) == 1) //target is SMAP
 				comStoreHndlr(addr, val, type); //write to real SPEED.
			//return;
			break;
		case 0x28: //	SPD_R_INTR_STAT
			//is there any effect on write?
			comStoreHndlr(addr, (val & 0xFFFC), type); //write to real SPEED.
			//return;
			break;
		case 0x2A: //	SPD_R_INTR_MASK
			updIntr(intrMask ^ val);
			intrMask = val & 0x3;
			comStoreHndlr(addr, (val & 0xFFFC), type); //write to real SPEED.
			//return;
			break; 
		case 0x32: //	SPD_R_XFR_CTRL
			//all bits but 0 and 7 seem write-only or read-only.
			if (val & 0x80) { //DMA start bit. Enables the SPEED making DMA requests (DREQ) to the IOP.
				intrStat |= 0x02; //fake that dma fifo has data
				p2mRunDma();
			} else {
				if (val & 0x01) { //DMA direction	
					//the direction should change reg 0x38 (and others)
					//1=write: IOP->SPEED->HDD; reg 0x38 decrements from 0x10 
					//For now just assume the counter (reg 0x38) will always be at 0 when this reg is being written.
					//There should be a counter and flags for the (imaginary) FIFO, which will be mirrorred by reg 0x38 in different ways for read and write, but there is no point in going that far.
					dbufStat = 0x50; //depends on fifo direction.
				} else {
					dbufStat = 0x60;	
				}
			}
			xfrCtrl = val & 0x81;
			break;
	/*	case 0x34: //	Total number of bytes, read from SPEED
//For now don't emulate these, even though PS2Linux uses them ...
			break;
		case 0x36: //	Total number of bytes, writen to SPEED

			break; */
		case 0x38: //	SPD_R_DBUF_STAT
			//if (val & 0x01) //reset something

			if (val & 0x02) {//reset FIFO
				if ((xfrCtrl & 0x01) != 0) //1=write: IOP->SPEED->HDD; reg 0x38 decrements from 0x10 
					dbufStat = 0x50; //depends on fifo direction.
				else
					dbufStat = 0x60;
			}
			break;
		//skipping the 0x60,... - weird registers.

	/*	case 0x62: //	State of (Ultra ?) DMA transfer and lines. read-only
			return ataDmaStat;
			break; */
		case 0x64: //	SPD_R_IF_CTRL
			tmp = val ^ ifCtrl; //a bit in tmp will get set only when it was changed.  
			ifCtrl = val & 0x4F; //0xCF;	//don't set the reset bit.
	//		if (val & 0x80) { //reset controller
	//			spdAtaCReset();
	//			break;
	//		}
			if (tmp & 0x40) //reset ATA HDD
				ataHddReset(val & 0x40); //if the arg. is !=0 => don't reset / 0=reset
			//if (tmp & 0x04) //SPD_IF_DMA_ENABLE	Enable SPEED acknowledging DMA requests from HDD.
				if (val & 0x04) { //was changed from inact to act => start stalled transfer if one was set-up
					//To make it safe, only a single sector will be transferred by DMA at a time, reg. 0x38 will (should) always contain 0x01 as count. This way the driver will always set-up transfer for only a single sector.
					dbufStat = 0x21;
					intrStat |= 0x02; //fake dma fifo has data
					causeIntr(intrMask); //1); //cause the DMA data ready interrupt.
					p2mRunDma();
				}
			//There actually isn't a need to check for change, because if it was initially set (the flag), then DMA would most likely already have been started (and it wouldn't wait for this reg. to change state.
			//no need to act on these here:
			//1	r/w DMA direction: 0=Write_to_HDD / 1=Read_from_HDD.
			//0	r/w 0=MDMA / 1=UDMA mode.		
			break;
		/*case 0x66: // r/w? it is writtable up to 0xFF and usually contains 0x1
			return spdReg0x66;
			break; */

//Might as well not emulate these:
/*		case 0x70: //	SPD_R_PIO_MODE
			return pioTiming;
			break;
		case 0x72: //	SPD_R_MWDMA_MODE
			return mwdmaTiming;
			break;
		case 0x74: //	SPD_R_UDMA_MODE
			return udmaTiming;
			break;
		case 0x76: //	CRC Initial value 
			return crcIn;
			break;
		case 0x78: //	CRC result 
			return crcOut;
			break; */

		default:
			if ((offs < 0x32) || (offs >= 0x80)) //don't access real hardware SPEED ATA regs.
				comStoreHndlr(addr, val, type); //common handler
	}
	}
	return;
}




